from slurpit.models.basemodel import BaseModel

class Device(BaseModel):
    """
    This class represents a network device with various attributes.

    Args:
        id (int): Unique identifier for the device.
        hostname (str): Hostname of the device.
        fqdn (str): Fully qualified domain name of the device.
        device_os (str): Operating system running on the device.
        disabled (int): Indicates whether the device is disabled (0 or 1).
        os_version (str, optional): Device OS version.
        serial (str, optional): Device Hardware Serial collected by SNMP.
        snmp_contact (str, optional): Device Contact collected by SNMP.
        snmp_description (str, optional): Device Description collected by SNMP.
        snmp_location (str, optional): Device Location collected by SNMP.
        snmp_uptime (str, optional): Device Uptime collected by SNMP.
        device_type (str, optional): Type of the device, such as 'router' or 'switch'.
        brand (str, optional): Brand of the device, e.g., 'Cisco'.
        added (str, optional): Date when the device was added to the system.
        last_seen (str, optional): Last date when the device was active or seen.
        port (int, optional): Network port number primarily used by the device.
        ipv4 (str, optional): IPv4 address assigned to the device.
        vault (str, readonly): Vault username.
        vault_id (int, optional): Configured vault id.
        site (str, optional): Site Name as created in Sites. When not defined site rules will apply.
        createddate (str, optional): The date the device record was created.
        changeddate (str, optional): The date the device record was last modified.
    """
        
    def __init__(
        self,
        id: int,
        hostname: str,
        fqdn: str,
        device_os: str,
        disabled: int,
        os_version: str = None,
        serial: str = None,
        snmp_contact: str = None,
        snmp_description: str = None,
        snmp_location: str = None,
        snmp_uptime: str = None,
        telnet:int = None,
        device_type: str = None,
        brand: str = None,
        added: str = None,
        last_seen: str = None,
        port: int = None,
        ipv4: str = None,
        vault: str = None,
        vault_id: int = None,
        site: str = None,
        createddate: str = None,
        changeddate: str = None,
    ):
        self.id = int(id)
        self.hostname = hostname
        self.fqdn = fqdn
        self.port = int(port) if port is not None else None
        self.ipv4 = ipv4
        self.device_os = device_os
        self.telnet = int(telnet) if telnet is not None else None
        self.disabled = int(disabled)
        self.os_version = os_version
        self.serial = serial
        self.snmp_contact = snmp_contact
        self.snmp_description = snmp_description
        self.snmp_location = snmp_location
        self.snmp_uptime = snmp_uptime
        self.device_type = device_type
        self.brand = brand
        self.added = added
        self.last_seen = last_seen
        self.vault = vault
        self.vault_id = vault_id
        self.site = site
        self.createddate = createddate
        self.changeddate = changeddate

class Vendor(BaseModel):
    """
    This class represents a vendor, defined primarily by the operating system and brand.

    Args:
        device_os (str): Operating system commonly associated with the vendor.
        brand (str): Brand name of the vendor, e.g., 'Apple' or 'Samsung'.
    """
    def __init__(
        self,
        device_os: str,
        brand: str,
        telnet: str = None
    ):
       
        self.device_os = device_os
        self.brand = brand
        self.telnet = telnet



from slurpit.models.basemodel import BaseModel

class User(BaseModel):
    """
    Represents a User model, inheriting from BaseModel, and defines attributes specific to a user.
    
    Args:
        id (int): Unique identifier for the user.
        first_name (str): User's first name.
        last_name (str): User's last name.
        email (str): User's email address.
        type (str): Type of user (e.g., admin, user).
        dark_mode (int): Indicates if dark mode is enabled (1) or not (0).
        createddate: The date when the user account was created. Initially set to None.
        changeddate: The date when the user account was last updated. Initially set to None.
    """

    def __init__(
        self,
        id: int,
        first_name: str,
        last_name: str,
        email: str,
        type: str,
        language: str,
        dark_mode: int,
        createddate: str = None,
        changeddate: str = None
    ):
        # Assigning the provided arguments to the instance variables
        self.id = int(id)  # Ensure id is stored as an integer
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.type = type
        self.language = language
        self.dark_mode = int(dark_mode)  # Ensure dark_mode is stored as an integer
        self.createddate = createddate
        self.changeddate = changeddate



from io import BytesIO
import httpx
try:
    import pandas as pd
except ImportError:
    pd = None

import functools
import inspect
import warnings

def json_to_csv_bytes(json_data):
    """
    Converts JSON data to CSV byte array.

    Args:
        json_data (list[dict]): A list of dictionaries representing JSON data.

    Returns:
        bytes: CSV formatted data as a byte array.
    """
    # Convert JSON to DataFrame
    df = pd.DataFrame(json_data)
    
    # Create a buffer
    buffer = BytesIO()
    
    # Convert DataFrame to CSV and save it to buffer
    df.to_csv(buffer, index=False)
    buffer.seek(0)  # Rewind the buffer to the beginning
    
    # Return bytes
    return buffer.getvalue()
        
def handle_response_data(response, object_class=None, export_csv=False, export_df=False):
    """
    Processes API response data dynamically, converting it to the requested format (CSV, DataFrame, or object list/raw data).
    Returns an empty list, empty DataFrame, or empty CSV byte string when the response payload is empty.

    Args:
        response (list[dict]): The raw response data from the API.
        object_class (type, optional): The class to instantiate for each item in the response. If None, raw data will be returned.
        export_csv (bool): If True, exports data as CSV.
        export_df (bool): If True, exports data as pandas DataFrame.

    Returns:
        list[object_class] | list[dict] | bytes | pd.DataFrame: List of object instances, raw data (list of dictionaries), CSV data, 
                                                               or DataFrame depending on the export flag.
    """
    if isinstance(response, httpx.Response):
        response = response.json()

    if not response:
        # Return empty objects based on export format
        if export_csv:
            return b''
        elif export_df:
            if pd:
                return pd.DataFrame()
            else:
                raise ImportError("Pandas is required for exporting data as a DataFrame.")
        else:
            return []

    if export_csv:
        return json_to_csv_bytes(response)
    elif export_df:
        if pd:
            return pd.DataFrame(response)
        else:
            raise ImportError("Pandas is required for exporting data as a DataFrame.")
    elif object_class:
        # Convert response to a list of object instances if an object class is provided
        return [object_class(**item) for item in response]
    else:
        # Return the raw data (list of dictionaries) if no object class is provided
        return response

string_types = (type(b''), type(u''))

def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter('always', DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2
                )
                warnings.simplefilter('default', DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2
            )
            warnings.simplefilter('default', DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


from slurpit.apis.baseapi import BaseAPI
from slurpit.models.scanner import Node
from slurpit.utils.utils import handle_response_data
import httpx

class ScannerAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initialize the ScannerAPI with the base URL of the API and an API key for authentication.
        
        Args:
            base_url (str): The root URL for the API endpoints.
            api_key (str): The API key used for authenticating requests.
            verfify (bool): Verify HTTPS Certificates.
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)  # Initialize the parent class with the API key

    async def get_nodes(self, batch_id: int, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of nodes based on the batch_id with pagination options and optionally exports the data to CSV format or pandas DataFrame.

        Args:
            batch_id (int): The batch identifier to filter nodes.
            offset (int): The starting index for pagination.
            limit (int): The maximum number of nodes to return.
            export_csv (bool): If True, returns the nodes data in CSV format as bytes.
            export_df (bool): If True, returns the nodes data as a pandas DataFrame.

        Returns:
            list[dict] | bytes | pd.DataFrame: A list of nodes, CSV data as bytes if export_csv is True, 
                                            or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/scanner/{batch_id}"
        response = await self._pager(url, offset=offset, limit=limit, paged_value="nodes")
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)

    async def get_finders(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of configured finders from the scanner API and optionally exports the data to CSV format or pandas DataFrame.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the finders data in CSV format as bytes.
            export_df (bool): If True, returns the finders data as a pandas DataFrame.

        Returns:
            list[dict] | bytes | pd.DataFrame: A list of finders, CSV data as bytes if export_csv is True,
                                            or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/scanner/configured/finders"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)

    async def get_crawlers(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of configured crawlers from the scanner API and optionally exports the data to CSV format or pandas DataFrame.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the crawlers data in CSV format as bytes.
            export_df (bool): If True, returns the crawlers data as a pandas DataFrame.

        Returns:
            list[dict] | bytes | pd.DataFrame: A list of crawlers, CSV data as bytes if export_csv is True,
                                            or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/scanner/configured/crawlers"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)

    async def start_scanner(self, scanner_data: dict):
        """
        Starts a scanning process with provided scanner configuration data.

        Args:
            scanner_data (dict): A dictionary containing the scanner configuration data. \n
                                The dictionary should include the following keys: \n
                                - "target" (str): The target to be scanned.
                                - "version" (str): The SNMP version, e.g., "snmpv2c".
                                - "batch_id" (int): The batch ID for the scanning process.
                                - "snmpv2c_key" (str): The SNMPv2c community string.
                                - "snmpv3_username" (str): The SNMPv3 username.
                                - "snmpv3_authkey" (str): The SNMPv3 authentication key.
                                - "snmpv3_authtype" (list of str): The SNMPv3 authentication types, possible values are ["none", "md5", "sha", "sha224", "sha256", "sha384", "sha512"].
                                - "snmpv3_privkey" (str): The SNMPv3 privacy key.
                                - "snmpv3_privtype" (list of str): The SNMPv3 privacy types, possible values are ["none", "des", "3des", "aes128", "aes192", "aes256", "aesblumenthal192", "aesblumenthal256"].

        Returns:
            dict: Status of the scanning process if successful.
        """
        url = f"{self.base_url}/scanner"
        response = await self.post(url, scanner_data)
        if response:
            scanner_status = response.json()
            return scanner_status

    async def clean_logging(self, datetime: str):
        """
        Triggers a cleaning process for scanner logs older than the specified datetime.

        Returns:
            dict: Result of the cleaning process if successful.
        """
        url = f"{self.base_url}/scanner/clean"
        request_data = {"datetime": datetime}
        response = await self.post(url, request_data)
        if response:
            clean_result = response.json()
            return clean_result

    async def get_status(self):
        """
        Retrieves the current status of the scanner.

        Returns:
            dict: The current status of the scanner if successful.
        """
        url = f"{self.base_url}/scanner/status"
        response = await self.get(url)
        if response:
            status_result = response.json()
            return status_result

    async def test_snmp(self, ip_data: dict):
        """
        Tests SNMP configuration by attempting to gather device information from the specified IP.

        Args:
            ip_data (dict): A dictionary containing the SNMP configuration details. \n
                            The dictionary should include the following keys: \n
                            - "ip" (str): The IP address of the device to connect to.
                            - "version" (str): The SNMP version, e.g., "snmpv2c".
                            - "snmpv2c_key" (str): The SNMPv2c community string.
                            - "snmpv3_username" (str): The SNMPv3 username.
                            - "snmpv3_authkey" (str): The SNMPv3 authentication key.
                            - "snmpv3_authtype" (list of str): The SNMPv3 authentication types, possible values are ["none", "md5", "sha", "sha224", "sha256", "sha384", "sha512"].
                            - "snmpv3_privkey" (str): The SNMPv3 privacy key.
                            - "snmpv3_privtype" (list of str): The SNMPv3 privacy types, possible values are ["none", "des", "3des", "aes128", "aes192", "aes256", "aesblumenthal192", "aesblumenthal256"].

        Returns:
            dict: Device information gathered via SNMP if successful.
        """
        url = f"{self.base_url}/scanner/test"
        timeout = httpx.Timeout(connect=10.0, read=60.0, write=60.0, pool=60.0)
        response = await self.post(url, ip_data, timeout=timeout)
        if response:
            device_info = response.json()
            return device_info

    async def get_queue_list(self, export_csv: bool = False, export_df: bool = False):
        """
        Gives a list of currently queued tasks for the scanner.

        Args:
            export_csv (bool): If True, returns the queued tasks data in CSV format as bytes. If False, returns list of queued tasks.
            export_df (bool): If True, returns the crawlers data as a pandas DataFrame.

        Returns:
            list[dict] | bytes: A list of queued tasks or CSV data as bytes if export_csv is True.
        """
        url = f"{self.base_url}/scanner/queue/list"
        response = await self.post(url, None)
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)
    
    async def clear_queue(self):
        """
        Clears the queue of the scanner by sending a DELETE request to the queue list endpoint.

        Returns:
            dict: The result of clearing the queue if successful.
        """
        url = f"{self.base_url}/scanner/queue/clear"
        response = await self.delete(url)
        if response:
            clear_result = response.json()
            return clear_result


from slurpit.apis.baseapi import BaseAPI
from slurpit.models.vault import Vault, SingleVault
from slurpit.utils.utils import handle_response_data

class VaultAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initializes a new instance of the VaultAPI class.

        Args:
            base_url (str): The base URL for the Vault API.
            api_key (str): The API key used for authentication.
            verfify (bool): Verify HTTPS Certificates.
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)  # Initializes the BaseAPI with the provided API key

    async def get_vaults(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Fetches a list of all vaults from the API and returns them as a list of Vault objects.
        Optionally exports the data to a CSV format or pandas DataFrame if specified.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the vault data in CSV format as bytes.
            export_df (bool): If True, returns the vault data as a pandas DataFrame.

        Returns:
            list[Vault] | bytes | pd.DataFrame: Returns a list of Vault objects if successful, bytes if exporting to CSV,
                                            or a pandas DataFrame if exporting to DataFrame.
        """
        url = f"{self.base_url}/vault"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, Vault, export_csv, export_df)
        
    async def get_vault(self, vault_id: int):
        """
        Fetches a single vault by its ID.

        Args:
            vault_id (int): The ID of the vault to retrieve.

        Returns:
            SingleVault: Returns a SingleVault object if the fetch is successful.
        """
        url = f"{self.base_url}/vault/{vault_id}"
        response = await self.get(url)
        if response:
            vault_data = response.json()
            return SingleVault(**vault_data)
    
    async def update_vault(self, vault_id: int, update_data: dict):
        """
        Updates a vault with specified data.

        Args:
            vault_id (int): The ID of the vault to update.
            update_data (dict): A dictionary containing the data to update in the vault. \n
                                The dictionary should include the following keys: \n
                                - "username" (str): The username for the vault.
                                - "password" (str): The password for the vault.
                                - "default" (int): A flag indicating whether this is the default vault (0 for no, 1 for yes).
                                - "device_os" (str): The operating system associated with the vault.
                                - "comment" (str): Any additional comments or notes about the vault.

        Returns:
            SingleVault: Returns a SingleVault object if the update is successful.
        """
        url = f"{self.base_url}/vault/{vault_id}"
        response = await self.put(url, update_data)
        if response:
            vault_data = response.json()
            return SingleVault(**vault_data)
    
    async def create_vault(self, new_vault: dict):
        """
        Creates a new vault with the specified data.

        Args:
            new_vault (dict): A dictionary containing the data for the new vault. \n
                            The dictionary should include the following keys: \n
                            - "username" (str): The username for the vault.
                            - "password" (str): The password for the vault.
                            - "default" (int): A flag indicating whether this is the default vault (0 for no, 1 for yes).
                            - "device_os" (str): The operating system associated with the vault.
                            - "comment" (str): Any additional comments or notes about the vault.

        Returns:
            SingleVault: Returns a SingleVault object if the creation is successful.
        """
        url = f"{self.base_url}/vault"
        response = await self.post(url, new_vault)
        if response:
            vault_data = response.json()
            return SingleVault(**vault_data)
    
    async def delete_vault(self, vault_id: int):
        """
        Deletes a vault by its ID.

        Args:
            vault_id (int): The ID of the vault to delete.

        Returns:
            SingleVault: Returns a SingleVault object if the deletion is successful.
        """
        url = f"{self.base_url}/vault/{vault_id}"
        response = await self.delete(url)
        if response:
            vault_data = response.json()
            return SingleVault(**vault_data)


from slurpit.models.basemodel import BaseModel

class Node(BaseModel):
    """
    This class represents a network device or node with various attributes that help identify and manage it in a network.

    Args:
        ip (str): The IP address of the network node, used for network communication and identification.
        device_type (str): The type of device, such as router, switch, or firewall, which helps in categorizing the node.
        hostname (str): The local name of the device, used for easier identification within a network.
        fqdn (str): The Fully Qualified Domain Name, providing a complete domain name address for the node.
        vendor (str): The manufacturer of the device, giving insights into the device's specifications and compatibility.
        product (str): The specific product name or model of the device as defined by the vendor.
        batch_id (int): A batch identifier that can link this node to a specific production or shipment batch.
        snmp_vars (str): A string representing SNMP variables used for monitoring the device via SNMP.
        timestamp (str): The date and time when the node data was last updated, important for tracking changes.
        error (str): A field to note any errors related to the node, useful for troubleshooting and maintenance.
    """
    def __init__(
        self,
        ip: str,
        device_type: str,
        hostname: str,
        fqdn: str,
        vendor: str,
        product: str,
        batch_id: int,
        snmp_vars: str,
        timestamp: str,
        error: str
    ):
        self.ip = ip
        self.device_type = device_type
        self.hostname = hostname
        self.fqdn = fqdn
        self.vendor = vendor
        self.product = product
        self.batch_id = batch_id
        self.snmp_vars = snmp_vars
        self.timestamp = timestamp
        self.error = error



from slurpit.apis.baseapi import BaseAPI
from slurpit.models.user import User
from slurpit.utils.utils import handle_response_data
class UserAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initializes a new instance of the UserAPI class, setting up the base URL and inheriting from BaseAPI.

        Args:
            base_url (str): The root URL for the user service API.
            api_key (str): The API key used for authorization in the API requests.
            verfify (bool): Verify HTTPS Certificates.
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)

    async def get_users(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Fetches a list of users from the API and returns them as a list of User objects.
        Optionally exports the data to a CSV format or pandas DataFrame if specified.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the user data in CSV format as bytes.
            export_df (bool): If True, returns the user data as a pandas DataFrame.

        Returns:
            list[User] | bytes | pd.DataFrame: Returns a list of User objects if successful, bytes if exporting to CSV,
                                            or a pandas DataFrame if exporting to DataFrame.
        """
        url = f"{self.base_url}/users"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, User, export_csv, export_df)

    async def get_user(self, user_id: int):
        """
        Fetches a single user by user ID from the API.

        Args:
            user_id (int): The unique identifier of the user to retrieve.

        Returns:
            User: Returns the User object if the fetch is successful.
        """
        url = f"{self.base_url}/users/{user_id}"
        response = await self.get(url)
        if response:
            user_data = response.json()
            return User(**user_data)

    async def update_user(self, user_id: int, update_data: dict):
        """
        Updates a user's information on the server.

        Args:
            user_id (int): The unique identifier of the user to update.
            update_data (dict): A dictionary containing the data to update. \n
                                The dictionary should include the following keys: \n
                                - "first_name" (str): The first name of the user.
                                - "last_name" (str): The last name of the user.
                                - "email" (str): The email address of the user.
                                - "language" (str): The preferred language of the user.
                                - "dark_mode" (int): The dark mode setting, where 0 indicates disabled and 1 indicates enabled.

        Returns:
            User: Returns the updated User object if the update is successful.
        """
        url = f"{self.base_url}/users/{user_id}"
        response = await self.put(url, update_data)
        if response:
            user_data = response.json()
            return User(**user_data)

    async def create_user(self, new_user: dict):
        """
        Creates a new user in the system.

        Args:
            new_user (dict): A dictionary containing the data of the new user to be created. \n
                            The dictionary should include the following keys: \n
                            - "first_name" (str): The first name of the user.
                            - "last_name" (str): The last name of the user.
                            - "email" (str): The email address of the user.
                            - "password" (str): The password for the user.
                            - "language" (str): The preferred language of the user.
                            - "dark_mode" (int): The dark mode setting, where 0 indicates disabled and 1 indicates enabled.

        Returns:
            User: Returns the newly created User object if the creation is successful.
        """
        url = f"{self.base_url}/users"
        response = await self.post(url, new_user)
        if response:
            user_data = response.json()
            return User(**user_data)

    async def delete_user(self, user_id: int):
        """
        Deletes a user from the system by user ID.

        Args:
            user_id (int): The unique identifier of the user to be deleted.

        Returns:
            User: Returns the User object if the deletion is confirmed.
        """
        url = f"{self.base_url}/users/{user_id}"
        response = await self.delete(url)
        if response:
            user_data = response.json()
            return User(**user_data)


from slurpit.models.basemodel import BaseModel

class Platform(BaseModel):
    """
    This class represent platform status.

    Args:
        status (str): The status of the platform, indicating its current operational state.
    """
    def __init__(self, status: str):
        self.status = status  # Store the platform's status

class PlatformLicense(BaseModel):
    """
    This class represents the Platform License details.

    Args:
        status (str): License status
        always_online (int): Does the system require an active internet connection
        amount_of_devices (int): Amount of devices that are licensed
        enterprise (int): Are enterprise features enabled
        license_end_date (str): The date the license ends
    """

    def __init__(
        self,
        status: str,
        always_online: int,
        amount_of_devices: int,
        enterprise: int,
        license_end_date: str
    ):
        self.status = status
        self.always_online = int(always_online)
        self.amount_of_devices = int(amount_of_devices)
        self.enterprise = int(enterprise)
        self.license_end_date = license_end_date


from slurpit.apis.baseapi import BaseAPI
from slurpit.models.platform import Platform, PlatformLicense
from slurpit.utils.utils import handle_response_data

class PlatformAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initializes a new instance of the PlatformAPI class, which extends BaseAPI. This class is designed to interact with platform-related endpoints of an API.

        Args:
            base_url (str): The root URL for the API endpoints.
            api_key (str): The API key used for authenticating requests.
            verfify (bool): Verify HTTPS Certificates.
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)

    async def license(self):
        """
        Retrieves a list of the current license status

        Returns:
            PlatformLicense: A PlatformLicense object initialized with the data from the API response if the request is successful.
        """
        url = f"{self.base_url}/platform/license" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return PlatformLicense(**platform_data)
        
    async def ping(self):
        """
        Sends a 'ping' request to the platform's ping endpoint to check connectivity and retrieve platform information.

        Returns:
            Platform: A Platform object initialized with the data from the API response if the request is successful.
        """
        url = f"{self.base_url}/platform/ping" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return Platform(**platform_data)
    
    async def version(self):
        """
        Returns the current installed Slurp'it version.

        Returns:
            dict: A dictionary initialized with the data from the Version response if the request is successful.
        """
        url = f"{self.base_url}/platform/version" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return platform_data  
    
    async def scraper_connectivity(self):
        """
        Test the Scraper (Data Collector) connectivity.

        Returns:
            Platform: A Platform object initialized with the data from the Service Status response if the request is successful (online/offline).
        """
        url = f"{self.base_url}/platform/scraper_connectivity" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return Platform(**platform_data)
    
    async def scanner_connectivity(self):
        """
        Test the Scanner (Device Finder) connectivity.

        Returns:
            Platform: A Platform object initialized with the data from the Service Status response if the request is successful (online/offline).
        """
        url = f"{self.base_url}/platform/scanner_connectivity" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return Platform(**platform_data)

    async def warehouse_connectivity(self):
        """
        Test the Warehouse connectivity.

        Returns:
            Platform: A Platform object initialized with the data from the Service Status response if the request is successful (online/offline).
        """
        url = f"{self.base_url}/platform/warehouse_connectivity" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return Platform(**platform_data)
    
    async def warehouse_db_size(self):
        """
        Get the Scraper (Data Collector) DB (MongoDB) size in MB

        Returns:
            Platform: A Platform object initialized with the data from the Warehouse DB Size response if the request is successful.
        """
        url = f"{self.base_url}/platform/warehouse_db_size" 
        response = await self.get(url)
        if response:
            platform_data = response.json()  
            return Platform(**platform_data)  


from slurpit.apis.baseapi import BaseAPI
from slurpit.models.device import Device, Vendor
from slurpit.utils.utils import handle_response_data, deprecated
import httpx

class DeviceAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initializes a new instance of DeviceAPI, extending the BaseAPI class.
        Sets up the base URL for API calls specific to devices and initializes authentication.

        Args:
            base_url (str): The root URL for the device-related API endpoints.
            api_key (str): The API key used for authenticating requests.
            verfify (bool): Verify HTTPS Certificates.
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)

    async def get_devices(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Fetches a list of devices from the API with pagination and optionally exports the data to a CSV format or pandas DataFrame.
        
        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the device data in CSV format as bytes.
                            If False, returns a list of Device objects.
            export_df (bool): If True, returns the device data as a pandas DataFrame.

        Returns:
            list[Device] | bytes | pd.DataFrame: A list of Device instances, CSV data as bytes if export_csv is True, 
                                                or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/devices"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, Device, export_csv, export_df)

    async def get_device(self, device_id: int):
        """
        Fetches a single device by its ID.

        Args:
            device_id (int): The unique identifier of the device to retrieve.

        Returns:
            Device: A Device instance if successful.
        """
        url = f"{self.base_url}/devices/{device_id}"
        response = await self.get(url)
        if response:
            device_data = response.json()
            return Device(**device_data)

    async def update_device(self, device_id: int, update_data: dict):
        """
        Updates a specific device using its ID.

        Args:
            device_id (int): The unique identifier of the device to update.
            update_data (dict): A dictionary containing the updated device attributes. \n
                            The dictionary should include the following keys: \n
                            - "hostname" (str): The hostname of the device.
                            - "fqdn" (str): The fully qualified domain name of the device.
                            - "port" (int): The port number for device communication, typically 22.
                            - "ipv4" (str): The IPv4 address of the device.
                            - "device_os" (str): The operating system running on the device.
                            - "device_type" (str): The type of the device.
                            - "disabled" (int): Status flag where 0 indicates enabled and 1 indicates disabled.

        Returns:
            Device: An updated Device instance if successful.
        """
        url = f"{self.base_url}/devices/{device_id}"
        response = await self.put(url, update_data)
        if response:
            device_data = response.json()
            return Device(**device_data)

    async def create_device(self, device_data: dict):
        """
        Creates a new device in the system.

        Args:
            device_data (dict): A dictionary containing the device attributes. \n
                            The dictionary should include the following keys: \n
                            - "hostname" (str): The hostname of the device.
                            - "fqdn" (str): The fully qualified domain name of the device.
                            - "port" (int): The port number for device communication, typically 22.
                            - "ipv4" (str): The IPv4 address of the device.
                            - "device_os" (str): The operating system running on the device.
                            - "device_type" (str): The type of the device.
                            - "disabled" (int): Status flag where 0 indicates enabled and 1 indicates disabled.

        Returns:
            Device: A newly created Device instance if successful.
        """
        url = f"{self.base_url}/devices"
        response = await self.post(url, device_data)
        if response:
            device_data = response.json()
            return Device(**device_data)

    async def delete_device(self, device_id: int):
        """
        Deletes a device using its ID.

        Args:
            device_id (int): The unique identifier of the device to delete.

        Returns:
            Device: A Device instance representing the deleted device if successful.
        """
        url = f"{self.base_url}/devices/{device_id}"
        response = await self.delete(url)
        if response:
            device_data = response.json()
            return Device(**device_data)

    async def sync_device(self, sync_data: dict):
        """
        Synchronizes a device with the given device data. Inserts or updates a device with the provided data.

        Args:
            sync_data (dict): A dictionary containing the device attributes to be synchronized. \n
                            The dictionary should include the following keys: \n
                            - "hostname" (str): The hostname of the device.
                            - "fqdn" (str): The fully qualified domain name of the device.
                            - "port" (int): The port number for device communication, typically 22.
                            - "ipv4" (str): The IPv4 address of the device.
                            - "device_os" (str): The operating system running on the device.
                            - "device_type" (str): The type of the device.
                            - "disabled" (int): Status flag where 0 indicates enabled and 1 indicates disabled.

        Returns:
            None | Device: None if the synchronization is successful and 'success' is reported by the API, 
                        otherwise returns a Device instance based on the returned data.
        """
        url = f"{self.base_url}/devices/sync"
        response = await self.post(url, sync_data)
        if response:
            sync_result = response.json()
            if 'success' in sync_result:
                return None

            return Device(**sync_result)

    async def get_vendors(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of vendors from the API and optionally exports the data to a CSV format or pandas DataFrame.
        
        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the vendor data in CSV format as bytes.
                            If False, returns a list of Vendor objects.
            export_df (bool): If True, returns the vendor data as a pandas DataFrame.

        Returns:
            list[Vendor] | bytes | pd.DataFrame: A list of Vendor instances, CSV data as bytes if export_csv is True, 
                                                or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/devices/vendors"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, Vendor, export_csv, export_df)

    async def get_types(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of device types from the API and optionally exports the data to CSV format or pandas DataFrame.
        
        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the device type data in CSV format as bytes.
                            If False, returns a list of device types as strings.
            export_df (bool): If True, returns the device type data as a pandas DataFrame.

        Returns:
            list[str] | bytes | pd.DataFrame: A list of device types, CSV data as bytes if export_csv is True, 
                                            or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/devices/types"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)


    async def get_snapshots(self, hostname: str, export_csv: bool = False, export_df: bool = False):
        """
        Retrieve latest data for a given hostname for all plannings, optionally filtered by date,
        and optionally exports the data to a CSV format or pandas DataFrame.
        
        Args:
            hostname (str): The hostname of the device for which snapshots are required.
            export_csv (bool): If True, returns the snapshot data in CSV format as bytes.
                            If False, returns a list of snapshot data dictionaries.
            export_df (bool): If True, returns the snapshot data as a pandas DataFrame.

        Returns:
            list[dict] | bytes | pd.DataFrame: A list of snapshot data dictionaries, CSV data as bytes if export_csv is True,
                                            or a pandas DataFrame if export_df is True.
        """
        url = f"{self.base_url}/devices/snapshot/all/{hostname}"
        response = await self.get(url)
        return handle_response_data(response, export_csv=export_csv, export_df=export_df)

    async def get_snapshot(self, hostname: str, planning_id: int, batch_id: int):
        """
        Retrieves latest data for a given hostname, planning id and batch id.

        Args:
            hostname (str): The hostname of the device.
            planning_id (int): The planning ID associated with the snapshot.
            batch_id (int): Batch ID, if empty it will use the latest batch id

        Returns:
            dict: Snapshot data as a dictionary if successful.
        """
        url = f"{self.base_url}/devices/snapshot/single/{hostname}/{planning_id}"
        if batch_id:
            url += f"/{batch_id}"
        response = await self.get(url)
        if response:
            snapshot_data = response.json()
            return snapshot_data


    async def get_snapshot_batches(self, hostname: str, planning_id: int):
        """
        Retrieve batch id data for a given hostname.

        Args:
            hostname (str): The hostname of the device.
            planning_id (int): Planning id, if not given it returns data for all planning ids

        Returns:
            dict: Batch id data for a given hostname and planning id.

        """
        url = f"{self.base_url}/devices/snapshot/batches/{hostname}"
        if planning_id:
            url += f"/{planning_id}"
        response = await self.get(url)
        if response:
            snapshot_data = response.json()
            return snapshot_data

    @deprecated("Use test_login() instead")
    async def test_ssh(self, ssh_info: dict):
        """
        Tests SSH connectivity using the provided SSH information.

        Args:
            ssh_info (dict): A dictionary containing the SSH credentials and details. \n
                            The dictionary should include the following keys: \n
                            - "username" (str): The username for SSH login.
                            - "password" (str): The password for SSH login.
                            - "ip" (str): The IP address of the device to connect to.
                            - "port" (int): The port number for SSH communication, typically 22.
                            - "device_os" (str): The operating system running on the device.

        Returns:
            dict: SSH test response as a dictionary if successful.
        """
        url = f"{self.base_url}/devices/test/ssh"
        timeout = httpx.Timeout(connect=10.0, read=60.0, write=60.0, pool=60.0)
        response = await self.post(url, ssh_info, timeout=timeout)
        if response:
            ssh_response = response.json()
            return ssh_response

    async def test_login(self, login_info: dict):
        """
        Tests SSH connectivity using the provided SSH information.

        Args:
            login_info (dict): A dictionary containing the SSH credentials and details. \n
                            The dictionary should include the following keys: \n
                            - "username" (str): The username for SSH login.
                            - "password" (str): The password for SSH login.
                            - "ssh_key" (str): The ssh_key for SSH login.
                            - "ssh_passphrase" (str): The ssh_passphrase for SSH Key logins.
                            - "enable_password" (str): The Enable password.
                            - "host" (str): The IP address of the device to connect to.
                            - "port" (int): The port number for SSH communication, typically 22.
                            - "device_os" (str): The operating system running on the device.

        Returns:
            dict: SSH test response as a dictionary if successful.
        """
        url = f"{self.base_url}/devices/test/login"
        timeout = httpx.Timeout(connect=10.0, read=60.0, write=60.0, pool=60.0)
        response = await self.post(url, login_info, timeout=timeout)
        if response:
            ssh_response = response.json()
            return ssh_response
    
    async def test_reachability(self, host: str, port: int):
        """
        Validate if the device is reachable by the provided port.

        Args:
            - "host" (str): The IP address of the device to connect to.
            - "port" (int): The port number for SSH communication, typically 22 or 23.

        Returns:
            dict: Reachability test response as a dictionary if successful.
        """
        url = f"{self.base_url}/devices/test/reachability/{host}/{port}"
        timeout = httpx.Timeout(connect=10.0, read=60.0, write=60.0, pool=60.0)

        response = await self.get(url, timeout=timeout)
        if response:
            reachability_response = response.json()
            return reachability_response


import httpx
from io import BytesIO
import os

class BaseAPI:
    def __init__(self, api_key, verify=True):
        """
        Initializes a new instance of the BaseAPI class with an `httpx` client.
        Args:
            api_key (str): The API key used for authorization, included in the headers of all requests.
            verfify (bool): Verify HTTPS Certificates.
        """
        self.client = httpx.AsyncClient(timeout=30, verify=verify)  # Creates an async client to persist certain parameters across requests
        self.client.headers.update({
            'Authorization': f'Bearer {api_key}',  # Bearer token for authorization
            'Content-Type': 'application/json'     # Sets default content type to JSON for all requests
        })

    async def get(self, url, **kwargs):
        """
        Sends an async GET request to the specified URL using httpx.
        Args:
            url (str): The URL to send the GET request to.
            **kwargs: Arbitrary keyword arguments that are forwarded to the `httpx.get` method.
        Returns:
            httpx.Response: The response object if the request was successful.
        Raises:
            httpx.HTTPStatusError: For responses with HTTP error statuses.
            httpx.RequestError: For network-related issues.
        """
        response = await self.client.get(url, **kwargs)  # Sends a GET request
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(e.response.text)
            raise e
        return response

    async def post(self, url, data, **kwargs):
        """
        Sends an async POST request with JSON data to the specified URL using httpx.
        Args:
            url (str): The URL to send the POST request to.
            data (dict): The JSON data to send in the body of the POST request.
            **kwargs: Arbitrary keyword arguments that are forwarded to the `httpx.post` method.
        Returns:
            httpx.Response: The response object if the request was successful.
        """
        response = await self.client.post(url, json=data, **kwargs)
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(e.response.text)
            raise e
        return response

    async def put(self, url, data, **kwargs):
        """
        Sends an async PUT request with JSON data to the specified URL using httpx.
        Args:
            url (str): The URL to send the PUT request to.
            data (dict): The JSON data to send in the body of the PUT request.
            **kwargs: Arbitrary keyword arguments that are forwarded to the `httpx.put` method.
        Returns:
            httpx.Response: The response object if the request was successful.
        """
        response = await self.client.put(url, json=data, **kwargs)
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(e.response.text)
            raise e
        return response

    async def delete(self, url, **kwargs):
        """
        Sends an async DELETE request to the specified URL using httpx.
        Args:
            url (str): The URL to send the DELETE request to.
            **kwargs: Arbitrary keyword arguments that are forwarded to the `httpx.delete` method.
        Returns:
            httpx.Response: The response object if the request was successful.
        """
        response = await self.client.delete(url, **kwargs)
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(e.response.text)
            raise e
        return response
    
    def save_csv_bytes(self, byte_data, filename):
        """
        Saves CSV byte array data to a CSV file.

        Args:
            byte_data (bytes): CSV data in byte array format.
            filename (str): The filename to save the CSV file as.
        """
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory)
            
        # Open file in binary write mode and write byte data
        with open(filename, 'wb') as file:
            file.write(byte_data)
        return True

    async def _pager(self, url, payload=None, offset=0, limit=5000, method="GET", paged_value="rows"):
        params = {'limit': limit}
        if payload is None:
            payload = {}
        payload_copy = payload.copy()  # Prevent modifying original payload
        payload_copy.update(params)

        data = list()
        async def get_page(offset_value):
            payload_copy['offset'] = offset_value
            try:
                if method == "GET":
                    resp = await self.get(url, params=payload_copy)
                elif method == "POST":
                    resp = await self.post(url, payload_copy)
                else:
                    raise ValueError(f"Unsupported method: {method}")
                temp_data = resp.json()
                if paged_value in temp_data:
                    return temp_data.get(paged_value, [])
                else:
                    return temp_data
            except Exception as e:
                raise Exception(f"Error fetching page with offset {offset_value}: {str(e)}")

        # Fetch initial page
        new_page = await get_page(offset)
        if not isinstance(new_page, list):
            return new_page            
        data.extend(new_page)

        # Continue fetching until fewer results than limit are returned
        while limit == len(new_page):
            offset += limit
            new_page = await get_page(offset)
            data.extend(new_page)

        return data


        



from slurpit.models.basemodel import BaseModel

class Planning(BaseModel):
    """
    This class represents a planning entity with various attributes used in management systems.

    Args:
        id (int): Unique identifier for the planning entry.
        name (str): Name of the planning entry.
        comment (int): Numeric comment identifier, which could link to further details or logs.
        disabled (str): Indicates the status of the planning entry, whether it's active or disabled.
        columns (list): A list of column names or specifications used in the planning entry.
        createddate (str): The date when the planning entry was created.
        changeddate (str): The date when the planning entry was last updated or modified.
    """
    
    def __init__(
        self,
        id: int,
        name: str,
        comment: int,
        disabled: str,
        columns: list,
        slug: str,
        createddate: str = None,
        changeddate: str = None
    ):
        
        self.id = int(id)
        self.name = name
        self.comment = comment
        self.disabled = disabled
        self.columns = columns
        self.slug = slug
        self.createddate = createddate
        self.changeddate = changeddate



from slurpit.models.basemodel import BaseModel

class Site(BaseModel):
    """
    This class represents a site.

    Args:
        id (int): Unique identifier for the site.
        sitename (str): Name of the site.
        description (str): Description of the site.
        street (str): Street of the site.
        county (str): County (district) of the site.
        state (str): State of the site.
        number (str): Number of the site.
        zipcode (str): Zipcode of the site.
        city (str): City of the site.
        country (str): Country of the site.
        phonenumber (int): Phone number of the site.
        status (int): Status flag where 0 indicates enabled and 1 indicates disabled.
        longitude (str): Longitude of the site.
        latitude (str): Latitude of the site.
        siteRules (dict): List of all Site rules that apply to this site
        createddate (str): Date site was created.
        changeddate (str): Date site was updated.
    """

    def __init__(
        self,
        id: int,
        sitename: str,
        description: str,
        street: str,
        county: str,
        state: str,
        number: str,
        zipcode: str,
        city: str,
        country: str,
        phonenumber: int,
        status: int,
        longitude: str,
        latitude: str,
        createddate: str = None,
        changeddate: str = None,
        siteRules: dict = []
    ):
        self.id = int(id)
        self.sitename = sitename
        self.description = description
        self.street = street
        self.county = county
        self.state = state
        self.number = number
        self.zipcode = zipcode
        self.city = city
        self.country = country
        self.phonenumber = phonenumber
        self.status = status
        self.longitude = longitude
        self.latitude = latitude
        self.createddate = createddate
        self.changeddate = changeddate
        self.siteRules = siteRules


class SiteRule(BaseModel):
    """
    This class represents a site rule.

    Args:
        id (int): Unique identifier for the site rule.
        rule (str): The rule to be applied.
        applied_to (str): The element to which the rule is applied.
        rule_order (int): The order of the rule.
        site_id (int): The ID of the site to which the rule is applied.
        disabled (int): Whether the rule is disabled.
        createddate (str): The date the rule was created.
        changeddate (str): The date the rule was last updated.
    """

    def __init__(
        self,
        id: int,
        rule: str,
        applied_to: str,
        rule_order: int = 0,
        site_id: int = 0,
        disabled: int = 0,
        createddate: str = None,
        changeddate: str = None,
    ):
        self.id = int(id)
        self.rule = rule
        self.rule_order = rule_order
        self.site_id = site_id
        self.applied_to = applied_to
        self.disabled = disabled
        self.createddate = createddate
        self.changeddate = changeddate


from slurpit.apis.baseapi import BaseAPI
from slurpit.models.site import Site, SiteRule
from slurpit.utils.utils import handle_response_data

class SiteAPI(BaseAPI):
    def __init__(self, base_url, api_key, verify):
        """
        Initializes a new instance of SiteAPI, extending the BaseAPI class.
        Sets up the base URL for API calls specific to sites and initializes authentication.

        Args:
            base_url (str): The root URL for the site-related API endpoints.
            api_key (str): The API key used for authenticating requests.
            verfify (bool): Verify HTTPS Certificates.,
        """
        formatted_url = "{}/api".format(base_url if base_url[-1] != "/" else base_url[:-1])  # Format the base URL to ensure it ends with '/api'
        self.base_url = formatted_url  # Set the formatted base URL
        super().__init__(api_key, verify)

    async def get_sites(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of sites from the API.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the site data in CSV format as bytes.
                            If False, returns a list of Site objects.
            export_df (bool): If True, returns the site data as a pandas DataFrame.

        Returns:
            List[Site]: A list of Site objects if the request is successful.
        """
        url = f"{self.base_url}/sites"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, Site, export_csv, export_df)

    async def get_site(self, site_id: int):
        """
        Fetches a single site by its ID.

        Args:
            site_id (int): The unique identifier of the site to retrieve.

        Returns:
            Site: A Site instance if successful.
        """
        url = f"{self.base_url}/sites/{site_id}"
        response = await self.get(url)
        if response:
            site_data = response.json()
            return Site(**site_data)

    async def update_site(self, site_id: int, update_data: dict):
        """
        Updates a specific site using its ID.

        Args:
            site_id (int): The unique identifier of the site to update.
            update_data (dict): A dictionary containing the updated site attributes. \n
                            The dictionary should include the following keys: \n
                            - "sitename" (str): Name of the site.
                            - "description" (str): Description of the site.
                            - "street" (str): Street of the site.
                            - "street" (str): Street of the site.
                            - "county" (str): County (district) of the site.
                            - "state" (str): State of the site.
                            - "number" (str): Number of the site.
                            - "zipcode" (str): Zipcode of the site.
                            - "city" (str): City of the site.
                            - "country" (str): Country of the site.
                            - "phonenumber" (int): Phone number of the site.
                            - "status" (int): Status flag where 1 indicates enabled and 0 indicates disabled.
                            - "longitude" (str): Longitude of the site.
                            - "latitude" (str): Latitude of the site.

        Returns:
            Site: An updated Site instance if successful.
        """
        url = f"{self.base_url}/sites/{site_id}"
        response = await self.put(url, update_data)
        if response:
            site_data = response.json()
            return Site(**site_data)
        
    async def create_site(self, site_data: dict):
        """
        Creates a new site in the system.

        Args:
            site_data (dict): A dictionary containing the site attributes. \n
                            The dictionary should include the following keys: \n
                            - "sitename" (str): Name of the site.
                            - "description" (str): Description of the site.
                            - "street" (str): Street of the site.
                            - "county" (str): County (district) of the site.
                            - "state" (str): State of the site.
                            - "number" (str): Number of the site.
                            - "zipcode" (str): Zipcode of the site.
                            - "city" (str): City of the site.
                            - "country" (str): Country of the site.
                            - "phonenumber" (int): Phone number of the site.
                            - "status" (int): Status flag where 1 indicates enabled and 0 indicates disabled.
                            - "longitude" (str): Longitude of the site.
                            - "latitude" (str): Latitude of the site.

        Returns:
            Site: A newly created Site instance if successful.
        """
        url = f"{self.base_url}/sites"
        response = await self.post(url, site_data)
        if response:
            site_data = response.json()
            return Site(**site_data)
        
    async def delete_site(self, site_id: int):
        """
        Deletes a site using its ID.

        Args:
            site_id (int): The unique identifier of the site to delete.

        Returns:
            Site: A Site instance representing the deleted site if successful.
        """
        url = f"{self.base_url}/sites/{site_id}"
        response = await self.delete(url)
        if response:
            site_data = response.json()
            return Site(**site_data)

    async def get_siterules(self, offset: int = 0, limit: int = 1000, export_csv: bool = False, export_df: bool = False):
        """
        Retrieves a list of site rules from the API.

        Args:
            offset (int): The starting index of the records to fetch.
            limit (int): The maximum number of records to return in one call.
            export_csv (bool): If True, returns the site rule data in CSV format as bytes.
                            If False, returns a list of SiteRule objects.
            export_df (bool): If True, returns the site rule data as a pandas DataFrame.

        Returns:
            List[Site]: A list of Site objects if the request is successful.
        """
        url = f"{self.base_url}/sites/siterules"
        response = await self._pager(url, offset=offset, limit=limit)
        return handle_response_data(response, SiteRule, export_csv, export_df)

    async def create_siterule(self, siterule_data: dict):
        """
        Creates a new site rule in the system.

        Args:
            siterule_data (dict): A dictionary containing the site rule attributes. \n
                            The dictionary should include the following keys: \n
                            - "rule" (str): The rule to be applied.
                            - "applied_to" (str): The element to which the rule is applied.
                            - "rule_order" (int): The order of the rule.
                            - "site_id" (int): The ID of the site to which the rule is applied.
                            - "disabled" (int): Whether the rule is disabled.

        Returns:
            SiteRule: A newly created SiteRule instance if successful.
        """
        url = f"{self.base_url}/sites/siterules"
        response = await self.post(url, siterule_data)
        if response:
            siterule_data = response.json()
            return SiteRule(**siterule_data)
        
    async def delete_siterule(self, siterule_id: int):
        """
        Deletes a site rule using its ID.

        Args:
            siterule_id (int): The unique identifier of the site rule to delete.

        Returns:
            SiteRule: A SiteRule instance representing the deleted site rule if successful.
        """
        url = f"{self.base_url}/sites/siterules/{siterule_id}"
        response = await self.delete(url)
        if response:
            siterule_data = response.json()
            return SiteRule(**siterule_data)

    async def test_siterule(self, test_data: dict):
        """
        Tests a site rule.
        """
        url = f"{self.base_url}/sites/siterules/test"
        response = await self.post(url, test_data)
        if response:
            test_result = response.json()
            return test_result

    async def reset_devices(self):
        """
        Resets the devices for all sites.
        """
        url = f"{self.base_url}/sites/reset_devices"
        response = await self.post(url, {})
        if response:
            return response.json()
